SuperStrict

Module mpm.DebugStream
ModuleInfo "Version: 1.3"
ModuleInfo "Author: mpmxyz"

Import brl.Stream
Import brl.StandardIO

Rem
	bbdoc: wraps a stream to debug it
	returns: the debugging stream
	about: The TDebugStream prints every read and write operation to the console.
	This is useful if you want to check your network code or if you you've got trouble with you file reader/saver.
	Don't forget to name your streams if you are debugging more then one at once!
	If you don't want to use the console: Use the parameter "debuggingStream" to redirect the output to another stream.
	
	The Bytes manipulated by ReadBytes and WriteBytes are printed as hexadecimal numbers.
	maxBytes specifies the maximal number of bytes being printed.
	
	You are able to debug streams via a "debug:name::"/"debug::" url prefix.
EndRem
Function DebugStream:TDebugStream(stream:TStream,name:String="",maxBytes:Int=-1,debuggingStream:TStream=Null)
	If stream=Null Then Return Null
	Local newStream:TDebugStream=New TDebugStream
	newStream.SetStream(stream)
	newStream.SetDebuggingStream(debuggingStream)
	If name Then newStream.name=name
	If maxBytes>=0 Then newStream.maxBytes=maxBytes
	Return newStream
EndFunction

Rem
	bbdoc: the debugging stream wrapper
EndRem
Type TDebugStream Extends TStreamWrapper
	Field name:String="unknown"
	Field maxBytes:Int=64
	Field _isBusy:Int=False	'protection against redundant logs
	Field debuggingStream:TStream
	
	Method OnRead(typ:String,data:String)
		Output "~q"+name+"~q reads "+typ+": "+data
	EndMethod
	Method OnWrite(typ:String,data:String)
		Output "~q"+name+"~q writes "+typ+": "+data
	EndMethod
	Method OnSeek(position:Int,realPosition:Int,oldPosition:Int)
		If realPosition=-1
			Output "~q"+name+"~q is not seekable (requested: "+position+")"
		ElseIf realPosition<>position
			Output "~q"+name+"~q moved from "+oldPosition+" to "+realPosition+" (requested: "+position+")"
		Else
			Output "~q"+name+"~q moved from "+oldPosition+" to "+realPosition
		EndIf
	EndMethod
	
	Method SetDebuggingStream(stream:TStream)
		If stream
			debuggingStream=stream
		Else
			debuggingStream=StandardIOStream
		EndIf
	EndMethod
	Method SetName(name:String)
		Self.name=name
	EndMethod
	Method Output(text:String)
		debuggingStream.WriteLine(text)
		debuggingStream.Flush()
	EndMethod
	
	Method Seek:Int( position:Int )
		Local oldPosition:Int=_stream.Pos()
		Local realPosition:Int=_stream.Seek(position)
		OnSeek(position,realPosition,oldPosition)
		Return realPosition
	EndMethod
	
	Method ReadBytes:Int( buf:Byte Ptr,count:Int )
		Local realCount:Int=_stream.ReadBytes( buf,count )
		If Not _isBusy
			OnRead(realCount+" of "+count+" bytes",BytesToHex(buf,Min(count,maxBytes)))
		EndIf
		Return realCount
	EndMethod
	Method WriteBytes:Int( buf:Byte Ptr,count:Int )
		Local realCount:Int=_stream.WriteBytes( buf,count )
		If Not _isBusy
			OnWrite(realCount+" of "+count+" bytes",BytesToHex(buf,Min(count,maxBytes)))
		EndIf
		Return realCount
	EndMethod
	
	Method ReadByte:Int()
		_isBusy=True
		Local value:Byte=_stream.ReadByte()
		OnRead("byte",value)
		_isBusy=False
		Return value
	EndMethod
	Method WriteByte( n:Int )
		_isBusy=True
		OnWrite("byte",n)
		_stream.WriteByte n
		_isBusy=False
	EndMethod
	
	Method ReadShort:Int()
		_isBusy=True
		Local value:Short=_stream.ReadShort()
		OnRead("short",value)
		_isBusy=False
		Return value
	EndMethod
	Method WriteShort( n:Int )
		_isBusy=True
		DebugLog "~q"+name+"~q writes short: "+n
		OnWrite("short",n)
		_stream.WriteShort n
		_isBusy=False
	EndMethod
	
	Method ReadInt:Int()
		_isBusy=True
		Local value:Int=_stream.ReadInt()
		OnRead("int",value)
		_isBusy=False
		Return value
	EndMethod
	Method WriteInt( n:Int )
		_isBusy=True
		OnWrite("int",n)
		_stream.WriteInt n
		_isBusy=False
	EndMethod
	
	Method ReadLong:Long()
		_isBusy=True
		Local value:Long=_stream.ReadLong()
		OnRead("long",value)
		_isBusy=False
		Return value
	EndMethod
	Method WriteLong( n:Long )
		_isBusy=True
		OnWrite("long",n)
		_stream.WriteLong(n)
		_isBusy=False
	EndMethod
	
	Method ReadFloat:Float()
		_isBusy=True
		Local value:Float=_stream.ReadFloat()
		OnRead("float",value)
		_isBusy=False
		Return value
	EndMethod
	Method WriteFloat( n:Float )
		_isBusy=True
		OnWrite("float",n)
		_stream.WriteFloat n
		_isBusy=False
	EndMethod
	
	Method ReadDouble:Double()
		_isBusy=True
		Local value:Double=_stream.ReadDouble()
		OnRead("double",value)
		_isBusy=False
		Return value
	EndMethod
	
	Method WriteDouble( n:Double )
		_isBusy=True
		OnWrite("double",n)
		_stream.WriteDouble n
		_isBusy=False
	EndMethod
	
	Method ReadLine:String()
		_isBusy=True
		Local value:String=_stream.ReadLine()
		OnRead("line","~q"+value+"~q")
		_isBusy=False
		Return value
	EndMethod
	
	Method WriteLine:Int( t:String )
		_isBusy=True
		OnWrite("line","~q"+t+"~q")
		Local number:Int=_stream.WriteLine( t )
		_isBusy=False
		Return number
	EndMethod
	
	Method ReadString:String( n:Int )
		_isBusy=True
		Local value:String=_stream.ReadString(n)
		OnRead("string","~q"+value+"~q")
		_isBusy=False
		Return value
	EndMethod
	
	Method WriteString( t:String )
		_isBusy=True
		OnWrite("string","~q"+t+"~q")
		_stream.WriteString t
		_isBusy=False
	EndMethod
	
	Method ReadObject:Object()
		_isBusy=True
		Local value:Object=_stream.ReadObject()
		If value
			OnRead("object","~q"+value.ToString()+"~q")
		Else
			OnRead("object","Null")
		EndIf
		_isBusy=False
		Return value
	EndMethod
	
	Method WriteObject( obj:Object )
		_isBusy=True
		If obj
			OnWrite("object","~q"+obj.ToString()+"~q")
		Else
			OnWrite("object","Null")
		EndIf
		_stream.WriteObject obj
		_isBusy=False
	EndMethod
	
	Function BytesToHex:String(bytes:Byte Ptr,count:Int)
		If count<=0 Then Return ""
		Local mem:Byte Ptr=MemAlloc(count*3-1)	'one hex number -> 4 bits, spaces between every byte
		For Local i:Int=0 Until count
			mem[i*3]=NibbleToChar((bytes[i] & $F0) Shr 4)
			mem[i*3+1]=NibbleToChar(bytes[i] & $F)
			If i<>count-1
				mem[i*3+2]=" "[0]
			EndIf
		Next
		Local text:String=String.FromBytes(mem,count*3-1)
		MemFree(mem)
		Return text
	EndFunction
	Function NibbleToChar:Byte(nibble:Byte)
		If nibble<10
			Return "0"[0]+nibble
		Else
			Return "A"[0]+(nibble-$A)
		EndIf
	EndFunction
EndType

Private
New TDebugFactory
Type TDebugFactory Extends TStreamFactory
	Method CreateStream:TStream(url:Object,proto:String,path:String,readable:Int,writeable:Int)
		If proto[..6]="debug:" Or proto="debug"
			Local stream:TStream=OpenStream(path,readable,writeable)
			If stream
				Return DebugStream(stream,proto[6..])
			EndIf
		EndIf
	EndMethod
EndType
